#ifndef SI_ARITHMETIC
#define SI_ARITHMETIC

#include <string>
#include "myvector.h"

// the possible arithmetic types
enum SIarithtype{SI_integer,SI_real,SI_unset};

// exception for bad SIarithmetic method calls
class SIarithmeticErr {
  
  // errorcode=0    bad call to SIarithmetic constructor
  //                (variable is not legal)
  // errorcode=1    call to divided_by with an empty expression
  // errorcode=2    call to multiplied with an empty expression
  // errorcode=3    bad call to SIarithmetic constructor
  //                (SI_unset type)
  // errorcode=4    attempt to get I value when SIarithmetic object
  //                is not an integer
  // errorcode=5    attempt to get D value when SIarithmetic object
  //                is not  a double
  // errorcode=6    other kinds of error
  
 private: int errorcode;
 
 public:  void setCode(int toset);
 
 int getCode();

};


// this class is used to keep track of all the variables that
// occour in the arithmetic expression

class SIvariable {

 protected:

  std::string varname;       // the symbolic name of the variable
  SIarithtype type;          // the type of the variable
  
 public:
  
  //constructor
  SIvariable(void);
  
  // copy constructor
  SIvariable(const SIvariable &p);
  
  // assignement operator
  const SIvariable & operator = (const SIvariable &p);
  
  // sets the variable name to toset
  void setVar(char *toset);
  
  // sets the variable type
  void setType(SIarithtype toset);
  
  // returns a string representation of the variable
  char *getVar(void);
  
  // returns the variable type
  SIarithtype getType(void);
  
  // variable comparation methods

  unsigned int operator==(SIvariable two);
  
  unsigned int operator<(SIvariable two);
  
  unsigned int operator<=(SIvariable two);
  
  unsigned int operator>(SIvariable two);
  
  unsigned int operator>=(SIvariable two);

  // destructor
  ~SIvariable(void);
};


class SIarithmetic {

 private:
  
  // tells whether the variable is already in the list
  unsigned int varIn(SIvariable t);

  // arithmetic operators support methods

  void plus(std::string toadd);
  
  void minus(std::string tosub);
  
  void multiplied(std::string tomult);

  void divided_by(std::string todiv);

  void power(std::string topow);
  
  void power_exp(std::string topow);
  
  void min(std::string tocompare);
  
  void max(std::string tocompare);

  // tells whether the expression is empty
  unsigned int isEmpty(void);
  
 protected: 

  SIarithtype exprtype;
  
  // the arithmetic expression converted to string
  std::string expr;
  
  // arithmetic expression variable list
  myvector<SIvariable> varlist;
  
 public:
  
  // Creation of a void arithmetic expression
  SIarithmetic(void);
  
  // Creation of an arithmetic expression from an integer number
  SIarithmetic(long a);

  // Creation of an arithmetic expression from a real number
  SIarithmetic(float a);
  
  // Creation of an arithmetic expression from a real number in rational
  // format, e.g., A/B. Rational format is returned by Sicstus.
  SIarithmetic(char* rational);

  // Creation of an arithmetic expression from a variable with the specified type
  SIarithmetic(char *varname, SIarithtype type);
  
  // destructor
  ~SIarithmetic();

  // Tells whether the expression is a pure number
  unsigned int isNumber(void);
  
  // the type of variable (real or integer) must be provided along
  // with varname. This method raises an exception if varname
  // is not a legal variable, or if type is SI_unset
  // A call to this method destroys the entire expression and just
  // creates a new one constituted by variable varname
  void setVar(char *varname, SIarithtype type);
  
  // Creates an expression constituted by the integer constant a
  // A call to this method destroys the entire expression and just
  // creates a new one constituted by the integer constant
  void setLong(long a);
  
  // Creates an expression constituted by the float constant a
  // A call to this method destroys the entire expression and just
  // creates a new one constituted by the float constant
  void setDouble(double a);
  
  // sets the arithmetic expression to the passed string
  // warning: this method can build invalid arithmetic expressions
  // if not properly used
  void setExpr(char *toset);
  
  // this method clears the variable list of the arithmetic expression
  void clearVariables(void);
  
  // adds to the arithmetic expression the passed variable
  // warning: this method can build invalid arithmetic expressions
  // if not properly used
  void addVariable(SIvariable toadd);

  // Arithmetic operators

  void plus(SIarithmetic p);

  void minus(SIarithmetic p);

  void multiplied(SIarithmetic p);

  void divided_by(SIarithmetic p);

  void power(SIarithmetic p);

  void power_exp(SIarithmetic p);

  void min(SIarithmetic p);

  void max(SIarithmetic p);

  void negate();

  void abs();

  void sin();

  void cos();

  void tan();

  // Comparison operator
  unsigned int operator ==(SIarithmetic p);

  // returns the type of the arithmetic expression
  SIarithtype getType();

  // returns all the variables contained in the arithmetic expression
  // in a list
  myvector<SIvariable> getVariables(void);

  // returns any positive integer if the arithmetic expression is
  // a mixed expression, i.e. if it contains both integer and
  // real variables; returns 0 otherwise
  unsigned int mixedExpr(void);

  // replaces all occurrences of the variable whose name is in oldvar
  // with the arithmetic expression newexpr
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char* oldvar,SIarithmetic newexpr);

  // replaces all occurrences of the variable whose name is in oldvar
  // with the variable name contained in newvar
  // if newvar is not a number, it is added to the current variables list with
  // the same type as the replaced variable(oldvar)
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char* oldvar,char* newvar);

  // replaces only the first occurrence of the variable whose name is in oldvar
  // with the variable name contained in newvar;
  // if newvar is not a number, it is added to the current variables list with
  // the same type as the replaced variable(oldvar)
  // returns a positive integer if at least one occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replaceOnce(char* oldvar,char* newvar);

  // returns the occurrences of the variable whose name is tosearch in
  // the current arithmetic expression
  unsigned int occurrences(char* tosearch);

  // replaces all occourrences of the variable whose name is in oldvar
  // with the variable name contained in newvar
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  // This variant of the method doesn't add the new variable to
  // variable list even though substitutions have been performed 
  unsigned int replaceNoAdd(char* oldvar,char* newvar);

  // returns a string representation of the arithmetic expression
  char* toString(void);
  char* toStringWithoutType(void);

  // copy constructor
  SIarithmetic(const SIarithmetic &p);   // deep copy 

  // assignement operator
  const SIarithmetic & operator = (const SIarithmetic &p);

  // puts the arithmetic expression inside parenthesis
  void prts(void);

  // tells wheter the expression is constituted by a single variable
  unsigned int singleVariable(void);

  // returns the integer value of the expression
  // throws an exception if the expression is not an integer
  long getLValue();
  
  // returns the double value of the expression
  // throws an exception if the expression is not a double
  double getDValue();
  
};

#endif
