#ifndef SI_FORMULA
#define SI_FORMULA

#include "SIatom.h"
#include "SIarithmetic.h"
#include "myvector.h"

#include <map>
#include <vector>
#include <string>

// Logical operator type definition
enum SI_logicalop{SI_AND,SI_OR,SI_NOT};



/******************************** atomlist CLASS *********************************/

// An atom list is an AND conjunction of atoms
// Every atom list has an algebric representation as a set of atoms, semantically
// associated to the logical AND conjunction of all the elements of the set

class atomlist {

 protected: 

  // the atoms composing the list  
  myvector<SIatom> list;
       
 public:
       
  ~atomlist(void);
  
  // Cuts the last atom of the list
  void cut(void);
  
  // Adds toadd to the list
  void add(SIatom toadd);
  
  // Resets the list the cursor to the first atom
  void resetCursor(void);
  
  // Returns the atom currently pointed by cursor
  SIatom& cursorElement(void);
  
  // Increases cursor by one position
  void incCurs(void);
  
  // Returns the number of elements of the list
  int length(void);

  // Returns the variables list of the atom list  
  myvector<SIvariable> getVariables(void);
  
  // Set strict inclusion operator
  unsigned int operator < (atomlist p);
  
  // Set general inclusion operator
  unsigned int operator <= (atomlist p);
  
  // Set strict contains operator
  unsigned int operator > (atomlist p);
  
  // Set general contains operator
  unsigned int operator >= (atomlist p);
  
  // Set equality operator
  unsigned int operator == (atomlist p);
  
  // copy constructor
  atomlist(atomlist const &p);
  
  // assignement operator
  const atomlist & operator = (const atomlist &p);
  
  // default constructor
  atomlist(void);
  
  // concatenates two atom lists
  void concatenate(atomlist p);
  
  // represents the atom list as a string
  char* writeList(void);

  // returns the number of occurrences of the variable whose name is tosearch
  // in the current atomlist
  unsigned int occurrences(char *tosearch);
  
};


/***************************** END OF atomlist CLASS *****************************/





/******************************* dnf_vector class ********************************/


// A DNF vector is an OR addition of atom lists
// It can be used to store formulas in their disjunctive normal form (DNF) 

class dnf_vector {

 protected:
  
  // The atom lists composing the DNF vector
  myvector<atomlist> vector;
  
  // concatenates two dnf_vectors
  dnf_vector concatenate(dnf_vector one,dnf_vector two);

 public:
  
  ~dnf_vector(void);
  
  dnf_vector(void);
  
  // Inserts toadd in the dnf vector
  void insert(atomlist toadd);   
  
  // copy constructor
  dnf_vector(dnf_vector const &p);
  
  // Inserts toadd in the dnf vector
  void add(atomlist toadd);
  
  // Reset the dnf vector cursor to position 0
  void resetCursor(void);
  
  // Erases the last atom list
  void cut(void);
  
  // Erases the atom list at position i
  void cut(int i);
  
  // Returns the current cursor position
  int curPos(void);

  // returns the variables list of the dnf vector 
  myvector<SIvariable> getVariables(void);
  
  // assignement operator
  const dnf_vector & operator = (const dnf_vector &p);
  
  // index operator
  atomlist operator [] (const int index);
  
  // returns the atom list currently pointed by cursor
  atomlist& cursorElement(void);
  
  // Increases cursor position by one
  void incCurs(void);
  
  // Returns the number of atom lists composing the dnf vector
  int length(void);
  
  // Erases the whole dnf vector
  void clean(void);
  
  // Transforms the dnf vector into a string
  char* writeVect(void);
  
  // concatenates the current dnf vector with p
  dnf_vector operator + (dnf_vector p);
  
  // algebraic product of current dnf vector with p
  dnf_vector operator * (dnf_vector p);
  
};



/*************************** END OF dnf_vector class ***************************/





/******************************** nodeErr class *******************************/


// The nodeErr class is used to treat exceptions deriving for bad calls
// of SIformula methods


class nodeErr {
 protected:
  
  // the errorcodes:
  // 0: call to methods for logical operators when the node is an atom
  // 1: call to methods for atoms when the node is a logical operator
  // 2: call to formulanode methods when the node is still unset
  // 3: attempt to set a non-atom node to the next atom pointer
  // 4: attempt to set an already set node
  // 5: other kinds of error
  int errcode;

 public:
  
  void setCode(int toset);
  
  int getCode(void);
  
};



/***************************** END OF nodeErr class *************************/




/****************************** formulanode class **************************/


// formulanode class is used to store  a node in the formula tree structure
// a formulanode can either represent an atom, or a logical operator
// only leaf nodes can represent atoms

class formulanode {

  friend class SIformula;
  friend class SIsimplifier;

 protected:
 
  // tells whether the node is an atom
  unsigned int isatom;

  // tells whether the node is a logical operator
  unsigned int logicop;
  
  // the dnf representation of the formula (or subformula) starting from the
  // current node
  dnf_vector dnform;
  
  // pointers to the children nodes, if the node is a logical operator
  myvector<formulanode*> children;

  // pointer to the next atom, if the node is an atom
  formulanode* nextatom;
  
  // the operator, if the node is a logical operator
  SI_logicalop opr;
  
  // the atom, if the node is an atom
  SIatom atm;
  
  // rebuilds the atom sequence on a tree
  formulanode* setNext(formulanode* next);

 public:
  
  formulanode(void);
  
  // A call to this method sets the node to be an tom
  // It raises an exception if one tries to set to atom a node that
  // has already been defined
  void setAtom(SIatom toset);
  
  // Sets the pointer to the next atom in the tree
  void setNextAtom(formulanode *toset);
  
  // A call to this method creates a logical operator with 
  // the passed dnf vector
  void setLogicalOp(SI_logicalop op,dnf_vector dnf);
  
  // adds a child to the tree
  void addChild(formulanode *toadd);
  
  // destructor of the tree
  ~formulanode(void);
	
  // copy constructor: deep copy
  // WARNING: this copy constructor doesn't replicate the atom
  // sequence on the tree leaves; the SIformula copy constructor does
  formulanode(formulanode &x);
  
  
  // adjusts all the dnf of the nodes of the tree
  // replacing oldvar with newvar
  void adjust(char* oldvar,char* newvar);

  // adjusts all the dnf of the nodes of the tree
  // replacing oldvar with newexpr
  void adjust(char* oldvar,SIarithmetic newexpr);

};

/***************************** END OF formulanode class ***************************/




/********************************* SIformula class *******************************/


// The SIformula class is used to store arithmetic-logic formulas, manipulate and
// manage them

class SIformula {

  friend class SIsimplifier;
  
 private:
  
  // Indicates the last atom of the formula (last leaf)
  formulanode* lastAtom(formulanode* p);
  
  // Indicates the first atom of the formula (first leaf)
  formulanode* firstAtom(formulanode* p);
  
  // creates a string representation of the subtree starting from tw
  char* writeNode(formulanode *tw);


 protected:

  // indicates whether the formula is identically true
  unsigned int istrue;

  // indicates whether the formula is identically false
  unsigned int isfalse;
  
  // pointer to the top node of the tree structure
  formulanode* top;

  // pointer to the first leaf of the tree structure
  formulanode* firstatom;

  // pointer to the last leaf of the tree structure
  formulanode* lastatom;
  
 public:

  // default constructor
  SIformula(void);

  // Constructor to create an identically false formula (if logicaval=0),
  // or an identically true formula (for all other possible values of logicval)
  SIformula(unsigned int logicval);
  
  // Creation of a formula constituted by a single atom p
  SIformula(SIatom p);
  
  // creates a formula starting from the passed atom
  // a call to this method destroys the instance to create a new one
  void setAtom(SIatom p);

  // creates a FALSE formula if logicval==0
  // creates a TRUE formula for all other possible values of logicval
  // a call to this method destroys the instance to create a new one
  void setLogicVal(unsigned int logicval);

  /*copy constructor*/
  SIformula(const SIformula &p);   /* deep copy */
    
  // assignement operator
  const SIformula & operator =(const SIformula &p);
  
  // logical AND operator  
  void SIand(SIformula p);
  
  // logical OR operator
  void SIor(SIformula p);
  
  // logical NOT operator
  void SInot();
  
  // logical IMPLIES operator
  void SIimplies(SIformula p);
  
  // creates a string representation of the current formula
  char* writeFormula(void);
  
  // replaces all occurrences of the variable whose name is in oldvar
  // with the arithmetic expression newexpr
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char *oldvar,SIarithmetic newexpr);
  
  // replaces all occurrences of the variable whose name is in oldvar
  // with the variable name contained in newvar
  // if instances of oldvar are found, newvar type will be the same type as
  // oldvar.
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char *oldvar,char *newvar);
  
  // destructor
  ~SIformula(void);
  
  // returns 0 if the instance is not an AND conjunction of equality 
  // atoms each with a single variable on the left, 
  // any other positive value if it is
  unsigned int eqConj(void);

  // returns all the variables contained in the formula in a list
  myvector<SIvariable> getVariables(void);
  
  // returns the pairs variable-value contained in the formula
  // precondition: the formula is a well-formed postcondition
  std::map<std::string,SIarithmetic*> getVarValues();

  //returns the atoms contained in the formula
  //no information is given about the structure
  std::vector<SIatom*> getAtoms();

  // returns any positive integer if the formula contains mixed 
  // atoms; returns 0 otherwise
  unsigned int mixedFormula(void);

  // tells whether the formula is identically false
  unsigned int isFalse(void);
  
  // tells whether the formula is identically true
  unsigned int isTrue(void);
  
  // returns a pointer to the first atom of the formula
  formulanode *firstAtom(void);
  
  // returns a pointer to the last atom of the formula
  formulanode *lastAtom(void);
  
  // returns the DNF of the current formula
  dnf_vector dnForm(void);
  
};


/**************************** END OF SIformula class *****************************/





/******************************* SUPPORT METHODS *********************************/



// replaces oldvar with newvar in the entire atomlist
// newvar is supposed to be belonging to the same type as oldvar
void replace(atomlist *t,char* oldvar,char* newvar);

// replaces the first occurrence of oldvar with newvar in the atomlist
// newvar is supposed to be belonging to the same type as oldvar
void replaceOnce(atomlist *t,char* oldvar,char* newvar);

// tells whether var is in tocontrol
unsigned int varIn(SIvariable var,myvector<SIvariable> tocontrol);



/**************************** END OF SUPPORT METHODS *****************************/



#endif
