#ifndef MYVECTOR
#define MYVECTOR



/*****************************************************
*  This class provides a dynamic vector data type.   *             
*  The STL vector is used to implement the structure.*
*                                                    *
*  Computational complexity analysis can be referred *
*  to the computational analysis of the STL vector   *
*  structure.                                        *
*                                                    *
*                                                    *
*  Exception handling:                               *
*                                                    *
*  Errors in index use are treated internally, and   *
*  cause the forced interruption of the C++ process  *
*  that originated them.                             *
*  Other kinds of erroneous use of the structure     *
*  should be implemented as exceptions and managed   *
*  outside this class.                               *
*                                                    *
*                                                    *
*                                                    *                     
*  To allow an efficient use of the vector a cursor  *
*  is provided. The module user can move through the *
*  dynamic structure by incrementing or decrementing *
*  the cursor.                                       *
*  The cursor position of an empty vector is -1      *
*                                                    *
*                                                    *
*****************************************************/


#include <stdlib.h>
#include <iostream>
#include <vector>

void SIcheckpoint();

template<class T> class myvector {

 private:

  std::vector<T> __myvector;
 
  int cursorposition; //indicates the cursor position in the structure
  
 public: 

  // inserts the element a at the bottom of the present object
  void Insert(T a) {
    if (__myvector.size()>0)
      __myvector.push_back(a);
    else {
      __myvector.push_back(a);
      cursorposition=0;
    }
    
  }
  
  // index operator
  T& operator [] (const int index){
    return __myvector[index];
  }
    
  
  // cuts the last element of the vector
  void Cut() {   
      
    switch (__myvector.size())
      {    
	// nothing to do if length=0
      case 0: break;
	// if length=1, only empties the list
      case 1: __myvector.pop_back();
	cursorposition=-1;
	break;
      default:
	if (cursorposition==__myvector.size()-1)
	  cursorposition--;
	__myvector.pop_back();
	break;
      }
  }
  
  // cuts the element indexed by i in the current object
  void Cut(int i) {

    // Treatment of index error 
    if ((i<0) || (i> (__myvector.size() - 1))) {
      std::cout << "\n" << "myvector object method cut(index): INVALID index value\n";
      exit(-1);
    } 


    switch (__myvector.size()){
      // nothing to do if length=0
    case 0: break;
      // if length=1, only empties the list
    case 1: cursorposition=-1;
      __myvector.pop_back();
      break;
      // if length >1 the list needs to be visited
    default:

      // cursorposition consistency preservation
      // in the following cases, the cursorposition variable must be decreased by one
      //
      // 1. i=0 and cursorposition=0
      // 2. both i and cursorposition indicate the last element
      // 3. i does not indicate the last element nor the first element, but
      //    it is less than cursorposition
      if (i==0) {
	if (cursorposition!=0)
	  cursorposition--;			 
      }
      else if (i==__myvector.size()-1) {
	if (cursorposition==__myvector.size()-1)
	  cursorposition--;
      }
      else {
	if (cursorposition>=i) 
	  cursorposition--;
      }


      // element removal procedure: the element is searched by visiting the vector
      // and it is removed.
      int temp_index=0;
      for (typename std::vector<T>::iterator cont=__myvector.begin(); cont!=__myvector.end();++cont) {
	if (temp_index==i) {
	  __myvector.erase(cont);
	  break;
	} 
	temp_index++;
      }
      break;
    }
  }
  
  // constructor 
  myvector<T>(void) {     
    cursorposition=-1;
  }
  
  // destructor: the default STL vector destructor is enough
  ~myvector<T>(void) {
  }
  
  // assignement operator; a is the vector assigned
  const myvector<T> & operator = (const myvector<T> &s) {
    __myvector=s.__myvector;
    cursorposition=s.cursorposition;
    return *this;
  }
    
  // copy constructor
  myvector<T>(const myvector<T> &s)
    : __myvector(s.__myvector), cursorposition(s.cursorposition) {}
    
  // returns the number of elements of myvector object
  int Length(void) {
    return __myvector.size();
  }
  
  // brings the cursor to point to the first element
  // cursor is set to NULL if the vector is empty
  void ResetCursor(void) {
    if (__myvector.size()==0) cursorposition=-1;
    else cursorposition=0;
  }
  
  // Increases by one the cursor position
  // Nothing is done if the operation is not possible
  void IncCurs(void) {
    if ( (__myvector.size()==0) || (cursorposition==__myvector.size()-1) ) return;
    else 
      cursorposition++;
  }
  
  // Return the element  currently pointed by cursor (modifiable)
  T& CursorElement(void) {

    // Index error: the caller tries to access the cursor element when the
    // vector is empty: this causes the forced interruption of execution
    if (__myvector.size()==0) {
      std::cout << "\nmyvector object method CursorElement(): invalid call.\n";
      std::cout << "Unrecoverable error, stopping execution.\n";
      SIcheckpoint();
      exit(-1);
    }

    return __myvector[cursorposition];
	  
  }
  
  // Returns the cursor position: -1 if the vector is empty
  int CurPos(void) {
    return cursorposition;
  }
  
};

#endif
