/**
 *Server for using the Sicstus simplifier via TCP/IP
 */

/**
 *Protocol (request/reply) for communicating with C++ SICSTUS Prolog

Commands:

I - Init. Initialize the state of Sicstus. The assumptions set used
    for proofs is initialized to the identically true assuption. Use
    immediately after connecting to the server. No other command will
    be accepted if a I command is not received.

F - Formula. Prefix for all commands related to transmitting formulas
    to the Sicstus server. Each formula transfer starts with a F1
    command and ends with a F0 command.

    F1 - Begin formula communication (formula operands and operators
         in postfixed format are expected hereafter).  As a
         precondition, there must be no other (part of) formula
         currently stored.

    FO - Operator. After this, the Sicstus server will wait and parse
         a string representing an operator.  Possible operators are +,
         -, *, /, _ (that is, unary negation of a number), >, G
         (greater than or equals), = (equals), !  (not equals), <, L
         (less than or equals), A (and), O (or), N (not), I (implies)
    
    FI - Integer value. After this, the Sicstus server will wait and
         parse a string representing an integer value.
    
    FR - Real value. After this, the Sicstus server will wait and
         parse a string representing a real value.
    
    FB - Boolean value. After this, the Sicstus server will wait and
         parse for a string representing a boolean value, i.e., "true"
         or "false"
    
    FS - Symbolic value. After this, the Sicstus server will wait and
         parse two subsequent strings: first, a single character
         representing the type of the symbol: i.e., I (integer), R
         (real) or B (boolean); second, a string representing the
         symbol name.

    F0 - End formula communication.

P - Prove. Prefix for commands that attempt to prove a formula, using
    the current assumption set as knowledge base. As a precondition,
    there must be a formula currently stored.

    PT - Attempt to prove that the formula is true. 

    PF - Attempt to prove that the formula is false.

A - Append. Add a formula to the current assumptions set and then
    discard it. As a precondition, there must be a formula currently
    stored.
  
    AT - Append the formula as it is.
  
    AF - Negate the formula and then append it.

D - Discard. Discard a formula. As a precondition, there must be a
    formula currently stored.

C - Clear. Deletes the whole assumption set.

Q - Quit. End the Sicstus session.



Return codes are as follows:

 250 Requested simplifier action okay, completed
 
 354 Continue with formula input; end with the F0 command

 451 Action completed with negative answer: unable to prove a formula
     or empty command

 554 Unrecoverable error, the transaction failed and the server as been
     reset: syntax errors, parsing errors, wrong command sequences, etc.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "SIInterface.h"

#define HALTONERROR(ret,mess) \
  if(ret<0){ \
  printf("*** halted on: %s\n",mess); \
  exit(0); \
  }

int main(int argc, char *argv[]){
  int s1,s2;
  int retval;
  int optval=1;
  struct sockaddr_in addr,addrclient;
  char buffer[256];
  
  printf("Welcome to simplifier interface server, prerelease.\n\n");
  s1=socket(AF_INET,SOCK_STREAM,0);
  setsockopt(s1,SOL_SOCKET,SO_REUSEADDR,(char *)&optval, sizeof(optval));
  HALTONERROR(s1,"**> socket creation: failed\n");
  printf("**> socket creation: succedeed\n");
  
  memset(&addr,0,sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(8888);
  retval=bind(s1,(struct sockaddr *)&addr,sizeof(struct sockaddr_in));
  HALTONERROR(retval,"**> bind: failed\n");
  printf("**> bind: succedeed\n");
  
//    len = sizeof(struct sockaddr_in);
//    retval = getsockname(s1,(struct sockaddr *)&addr,&len);
//    HALTONERROR(retval,"getsockname: failed\n");
//    printf("**> getsockname: succedeed - port: %d\n",ntohs(addr.sin_port));
  
  retval = listen(s1,3);
  HALTONERROR(retval,"**> listen: failed\n");
  printf("**> listen: succedeed\n");  
  printf("**> waiting for connections...\n");

  memset(&addrclient,0,sizeof(struct sockaddr_in));
  socklen_t len = sizeof(struct sockaddr_in);
  s2=accept(s1,(struct sockaddr *)&addrclient,&len);
  HALTONERROR(s2,"**> accept: failed\n");
  printf("**> accept: succedeed - client port: %d - client address: %ul\n",ntohs(addrclient.sin_port),addrclient.sin_addr.s_addr);
  
  SIInterface *simplifier = new SIInterface();
  char replyCode[256];
  const char *tmpCode;
  int receivedChars;
  printf("**> sicstus started...\n");
  do{
    receivedChars=recv(s2,buffer,255,0);
    buffer[receivedChars]='\0';
    HALTONERROR(receivedChars,"**> read: failed\n");
    printf("FROM CLIENT:%d: %s\n",receivedChars,buffer);
    
    if(receivedChars==0){
      strcpy(replyCode, "451 Requested action aborted: empty command\n");
      simplifier->reset();
    }
    else{
      switch(buffer[0]){
      case 'I': 
	if(receivedChars>1){
	  strcpy(replyCode, "554 Syntax error in a command: init\n");
	  simplifier->reset();
	}
	else{
	  tmpCode=simplifier->init();
	  strcpy(replyCode, tmpCode);
	}
	break;
      case 'F':
	if(receivedChars<2){
	  strcpy(replyCode, "554 Syntax error in a command: formula*\n");
	  simplifier->reset();
	}  
	else{
	  switch(buffer[1]){
	  case '1': tmpCode=simplifier->formulaStart(); break;
	  case 'O': tmpCode=simplifier->formulaOperator(buffer+2); break;
	  case 'I': tmpCode=simplifier->formulaInteger(buffer+2); break;
	  case 'R': tmpCode=simplifier->formulaReal(buffer+2); break;
	  case 'B': tmpCode=simplifier->formulaBoolean(buffer+2); break;
	  case 'S': tmpCode=simplifier->formulaSymbol(buffer+2); break;
	  case '0': tmpCode=simplifier->formulaFinalize(); break;
	  default: tmpCode=NULL;
	  }
	  if(!tmpCode){
	    strcpy(replyCode,
		   "554 Syntax error. Formula action does not exist\n");
	    simplifier->reset();
	  }
	  else strcpy(replyCode, tmpCode);
	  break;
	}
	break;
      case 'P':
	if(receivedChars!=2){
	  strcpy(replyCode, "554 Syntax error in a command: prove\n");
	  simplifier->reset();
	}  
	else{
	  switch(buffer[1]){
	  case 'T': tmpCode=simplifier->prove(true); break;
	  case 'F': tmpCode=simplifier->prove(false); break;
	  default: tmpCode=NULL;
	  }
	  if(!tmpCode){
	    strcpy(replyCode,
		   "554 Syntax error. Wrong truth value for a  proof\n");
	    simplifier->reset();
	  }
	  else strcpy(replyCode, tmpCode);
	}
	break;
      case 'A': 
	if(receivedChars!=2){
	  strcpy(replyCode, "554 Syntax error in a command: append\n");
	  simplifier->reset();
	}  
	else{
	  switch(buffer[1]){
	  case 'T': tmpCode=simplifier->append(true); break;
	  case 'F': tmpCode=simplifier->append(false); break;
	  default: tmpCode=NULL;
	  }
	  if(!tmpCode){
	    strcpy(replyCode,
		   "554 Syntax error. Wrong truth value in command: append\n");
	    simplifier->reset();
	  }
	  else strcpy(replyCode, tmpCode);
	}
	break;
      case 'D': 
	if(receivedChars>1){
	  strcpy(replyCode, "554 Syntax error in a command: discard\n");
	  simplifier->reset();
	}
	else{
	  tmpCode=simplifier->discard();
	  strcpy(replyCode, tmpCode);
	}
	break;
      case 'C': 
	if(receivedChars>1){
	  strcpy(replyCode, "554 Syntax error in a command: clear\n");
	  simplifier->reset();
	}
	else{
	  tmpCode=simplifier->clear();
	  strcpy(replyCode, tmpCode);
	}
	break;
      case 'Q': 
	strcpy(replyCode, "250 Connection closed as requested\n");
	delete simplifier;
	break;
      default: 
	strcpy(replyCode, "554 Syntax error: command does not exist\n");
	simplifier->reset();
	break;
      }
    }
    retval=send(s2,replyCode,strlen(replyCode),0);
    HALTONERROR(retval,"ERROR: failed send to client\n");
    printf("TO CLIENT: %s", replyCode);
    //trasmissione sul socked di replyCode!!!!
  }while(buffer[0]!='Q');
  shutdown(s2,2);
  shutdown(s1,2);
}
