/* This file contains the interface to pass query strings and receive back 
   answers from the prolog interpreter */
//#define TWO_DEBUG_MODE

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <iostream>
#include <fstream>
#include <ext/stdio_filebuf.h>
#include <string>
#include "myvector.h"
#include "utils.h"
#include "SicstusC.h"


/**************************** AUXILIARY METHODS *******************************/

unsigned int isMLetter(char t) {
  return ( (t>='a') && (t<='z') );
};

/****************** END OF AUXILIARY METHODS **********************************/




/************************** INTERFACE METHODS ********************************/


SicstusC::SicstusC(void) {
  initialized=0;
};

// Initializes the prolog engine searching for the communication pipes
// in the specified path (pipespath)
// return 0 if the initilization fails, 1 otherwise
// slot must  be 0 or 1, at most 2 prolog processes are allowed
unsigned int SicstusC::initialize(unsigned int slot) {

  std::string path,f1,f2,cmd,str,cmd1;
    
  // Creation of the communication pipes
  int pipeToSicstus[2], pipeFromSicstus[2];
  if (pipe(pipeToSicstus) == -1){
    std::cerr<<"ERROR: Unable to create a pipe with Sicstus Prolog\n";
    exit(1);
  }
  if (pipe(pipeFromSicstus) == -1){
    std::cerr<<"ERROR: Unable to create a pipe with Sicstus Prolog\n";
    exit(1);
  }

  // Forking an hosting process for Sicstus Prolog with redirected I/O 
  int pid = fork ();
  if(pid<0){
    std::cerr<<"ERROR: Unable to fork an hosting process for Sicstus Prolog\n";
    exit(1);
  }
  else if(pid==0){ //In the child process
    dup2(pipeToSicstus[0], fileno(stdin)); //Copy the read-end of the pipe
    dup2(pipeFromSicstus[1], fileno(stdout)); //Copy the write-end of the pipe
    dup2(pipeFromSicstus[1], fileno(stderr)); //Copy the write-end of the pipe
    
    //starting sicstus.exe
    cmd+="sicstus";
    char* cmdc=stringToChar(cmd);
    if(!system(cmdc)){
      std::cerr<<"ERROR: Unable to start SicstusProlog: " <<
	"Check if it is in the PATH\n";
      exit(-1);
    }
    //Sictsus exited at this point
    delete cmdc;
    exit(0);
  }//end of the child process


  // Attaching FILE* streams to the pipes for writing (sicstrus_in)
  // and reading (sicstus_out) Sicstus Prolog I/O

  FILE *sicstus_in, *sicstus_out;
  if ((sicstus_in = fdopen(pipeToSicstus[1], "w")) == NULL){
    std::cerr<<"ERROR: Unable to create a stream for writing to Sicstus Prolog\n";
    kill(pid,SIGKILL); //kill the child process 
    exit(1);
  }
  if ((sicstus_out = fdopen(pipeFromSicstus[0], "r")) == NULL){
    std::cerr<<"ERROR: Unable to create a stream for reading from Sicstus Prolog\n";
    kill(pid,SIGKILL); //kill the child process 
    exit(1);
  }

  // Wrapping c++ stream to the FILE*:
  // stream fin wraps sicstus_out and reads input from Sictsus Prolog  
  // stream fout wraps sicstus_in and write output from Sictsus Prolog  
  __gnu_cxx::stdio_filebuf<char> * ibuf = 
    new __gnu_cxx::stdio_filebuf<char>(sicstus_out,std::ios_base::in,1);
  __gnu_cxx::stdio_filebuf<char> * obuf = 
    new __gnu_cxx::stdio_filebuf<char>(sicstus_in,std::ios_base::out);
  fin.std::istream::rdbuf(ibuf);
  fout.std::ostream::rdbuf(obuf);

  //Getting the Sicstus Prolog two line header
  for(int cont=0;cont<2;cont++) {
    std::getline(fin,str);
    std::cout<< "FROM SIMPLIFIER: " << stringToChar(str) << std::endl;
  }
  

  // Configures the prolog process by loading the constraint logic programming
  // libraries. If some library is not loaded the initialization process fails
  fout << "use_module(library(clpq))." << std::endl << std::endl;
  fout.flush();
  std::cout<< "TO SIMPLIFIER: use_module(library(clpq))." << std::endl;
  str="";

  // NOTE: GIOVANNI 2005 NOV 12: Now (porting on cygwin, glib 3.3.3,
  // and Sicstus 3.12.3) it looks like I get carriage return
  // characters (\r) at the end of each line sent by Sicstus
  // Prolog. Hereafter, I am modifying all Sicstus Prolog answers,
  // cutting the last character. I am not happy with this workaround,
  // but at the moment I do not see any more portable solution.
  // Maintenance hint: this workaround impacts only this file!
  while(  (str!="yes") && (str!="no") ) {
    std::getline(fin,str);
    if(str[str.length()-1]=='\r')
      str=str.substr(0,str.length()-1); //see NOTE: GIOVANNI 2005 NOV 12
#ifdef TWO_DEBUG_MODE
    std::cout<<"FROM SIMPLIFIER:"<<str.length()<<": "<<str<<std::endl;
    std::cout.flush();
#endif
  }
  if(str=="no") {
      std::cout<<"ERROR: simplifier cannot load library clpfd."<<std::endl;
      return 0;
  }
  fout << "use_module(library(clpfd))." << std::endl << std::endl;
  fout.flush();
  std::cout<< "TO SIMPLIFIER: use_module(library(clpfd))." << std::endl;
  str="";
  while(  (str!="yes") && (str!="no") ) {
    std::getline(fin,str);
    if(str[str.length()-1]=='\r')
      str=str.substr(0,str.length()-1); //see NOTE: GIOVANNI 2005 NOV 12
#ifdef TWO_DEBUG_MODE
    std::cout<<"FROM SIMPLIFIER: "<<str.length()<<": "<<str<<std::endl;
    std::cout.flush();
#endif
  };
  if(str=="no") {
      std::cout<<"ERROR: simplifier cannot load library clpfd."<<std::endl;
      return 0;
  }
  fout << "prolog_flag(toplevel_print_options, _,[quoted(true),numbervars(true),portrayed(true),max_depth(0)])." << std::endl << std::endl;
  fout.flush();
  std::cout<< "TO SIMPLIFIER: prolog_flag(toplevel_print_options, _,[quoted(true),numbervars(true),portrayed(true),max_depth(0)])." << std::endl;
  str="";
  while(  (str!="yes") && (str!="no") ) {
    std::getline(fin,str);
    if(str[str.length()-1]=='\r')
      str=str.substr(0,str.length()-1); //see NOTE: GIOVANNI 2005 NOV 12
#ifdef TWO_DEBUG_MODE
    std::cout<<"FROM SIMPLIFIER:"<<str.length()<<": "<<str<<std::endl;
    std::cout.flush();
#endif
  }
  if(str=="no") {
      std::cout<<"ERROR: simplifier cannot set toplevel_print_options."<<std::endl;
      return 0;
  }

  initialized=1;
  return 1;
};


void SicstusC::closeProlog(void) {

  // Sends the halt command to Sicstus
  if(initialized) { 
    fout << "halt." << std::endl << std::endl;
#ifdef TWO_DEBUG_MODE
    std::cout<<"TO SIMPLIFIER: halt."<<std::endl;
    std::cout.flush();
#endif
    fout.flush();
    fout.close();
    fin.close();
    initialized=0;
  }

};

// returns in a myvector object all the lines that are the answer to the
// query string q

myvector<char*> SicstusC::query(char* q) {

  myvector<char*> toreturn;
  std::string str;

#ifdef TWO_DEBUG_MODE
  std::cout << "TO SIMPLIFIER: " << q << "\n";  
  std::cout.flush();
#endif  
  // Sends the query q to Sicstus and reads back the answer, eliminating 
  // Sicstus unnecessary substrings
  
  fout << q << std::endl << std::endl;
  fout.flush();

  while( (str!="yes") && (str!="no")) {
    std::getline(fin,str);
    if(str[str.length()-1]=='\r')
      str=str.substr(0,str.length()-1); //see NOTE: GIOVANNI 2005 NOV 12
#ifdef TWO_DEBUG_MODE
    std::cout<<"FROM SIMPLIFIER:"<<str.length()<<": "<<str<<std::endl;
    std::cout.flush();
#endif

    if(str.length()>=0) { 
      if (str.substr(0,5)=="clpfd") continue;
      if (str.substr(0,8) == "_Inf = 0,") continue; //If, present, exclude the constant result Inf=0 results for integer constrained queries, which use bb_Inf([...], 0, _Inf)
      std::string cp;
      if ((str=="no") || (str=="yes"))
	toreturn.Insert(stringToChar(str));
      else {
	for(int cont=0;cont<str.length();cont++) {
	  if( (str[cont]!='?') && 
	      (str[cont]!=',') && 
	      (str[cont]!=' ') && 
	      (str[cont]!='{') && 
	      (str[cont]!='}') && 
	      (!isMLetter(str[cont])) && 
	      (str[cont]!=':')  ) 
	    cp+=str[cont];
	}
	if(cp.length()>0) toreturn.Insert(stringToChar(cp));
      }
    }
  }

  return toreturn;
};

/*****************END OF INTERFACE METHODS *************************************/
